package at.acdp.urweb.rtde;

import at.acdp.urweb.rtde.packets.*;
import org.slf4j.LoggerFactory;

import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.net.Socket;
import java.util.HashMap;
import java.util.concurrent.Semaphore;

import static at.acdp.urweb.rtde.CommandType.*;

public class RTDEClient implements Runnable {
    private final static org.slf4j.Logger logger = LoggerFactory.getLogger(RTDEClient.class);
    private final Semaphore waitingRequestSem = new Semaphore(0);
    private final Semaphore socketOpen = new Semaphore(0);

    private volatile IRtdeData waitingRequest=null;

    public static int RTDE_PROTOCOL_VERSION = 2;
    private final String ip;
    private final int port;
    private final boolean _running = true;
    private DataOutputStream dos;
    private DataInputStream dis;
    private Thread readThread;
    private RtdeSetupOutputs outputs;
    private HashMap<String, String> lastData;


    public RTDEClient(String ip, int port) {
        this.ip = ip;
        this.port = port;
    }

    public void start() throws IOException, InterruptedException {
        readThread = new Thread(this);
        readThread.start();
        socketOpen.acquire();
        init();
    }

    public HashMap<String, String> getLastData() {
        return lastData;
    }

    public void request(IRtdeData data) throws IOException, InterruptedException {
        if(readThread.getId()==Thread.currentThread().getId()) {
            logger.error("request() must not be called from readThread.");
        }
        if(data.hasReply()) {
            waitingRequest=data;
        }
        data.send(dos);
        if(data.hasReply()) {
            waitingRequestSem.acquire();
            waitingRequest=null;
        }
        return;
    }

    // Internal method that actually reads the data
    private void readSocket() throws IOException, InterruptedException {
        int length = dis.readShort();
        int type=dis.readByte() & 0xFF;
        var wreq=waitingRequest;
        if(wreq != null && type==wreq.getType()) {
            waitingRequest.read(dis, length-3);
            if(type== RTDE_CONTROL_PACKAGE_SETUP_OUTPUTS) {
                outputs= (RtdeSetupOutputs) waitingRequest;
            }
            waitingRequestSem.release();
        } else switch(type) {
            case RTDE_TEXT_MESSAGE:
                var rpv = new RtdeTextMessage();
                rpv.read(dis, length-3);
                logger.info(String.format("RTDE_TEXT_MESSAGE: %s", rpv.text));
                break;
            case RTDE_DATA_PACKAGE:
                var p = new RtdeOutDataPackage(outputs);
                p.read(dis, length-3);
                lastData=p.getPdata();
                lastData.entrySet().forEach(entry->{
                    System.out.println(entry.getKey() + " " + entry.getValue());
                });
                break;

            default: {
                logger.warn(String.format("Unexpected package type: %d", type));
                byte[] buf=new byte[length];
                dis.readFully(buf);
            }
        }
    }

    @Override
    public void run() {
        try {
            Socket rt = new Socket(ip, port);
            rt.setSoTimeout(0);
            rt.setReuseAddress(true);
            rt.setTcpNoDelay(true);
            if (rt.isConnected()) {
                System.out.println("Connected to UR Realtime Client");
            }
            dos = new DataOutputStream(rt.getOutputStream());
            dis = new DataInputStream(rt.getInputStream());
            socketOpen.release();
            while (_running) {
                try {
                    readSocket();
                } catch (IOException | InterruptedException e) {
                    e.printStackTrace();
                }
            }
        } catch (IOException  e) {
            logger.warn("run",e);
        }
    }

    public void init() throws IOException, InterruptedException {
        var rpv = new RtdeRequestProtocolVersion();
        request(rpv);
        logger.info("RtdeRequestProtocolVersion succes:" + rpv.success);
        var ruv = new RtdeRequestURVersion();
        request(ruv);
        logger.info(String.format("RtdeRequestURVersion: %d.%d.%d.%d",ruv.major, ruv.minor, ruv.build, ruv.bugfix));
    }
}
